#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/socket.h>

#include <sys/time.h>
#include <sys/types.h>

#include "tsocket.h"
#include "thread.h"
#include "util.h"

static void make_async(int);

static void
make_async (int s)
{
  int n;

  /* Make file file descriptor nonblocking. */
  if ((n = fcntl (s, F_GETFL)) < 0
      || fcntl (s, F_SETFL, n | O_NONBLOCK) < 0)
    fatal ("O_NONBLOCK: %s\n", strerror (errno));

  /* You can pretty much ignore the rest of this function... */

  /* Many asynchronous programming errors occur only when slow peers
   * trigger short writes.  To simulate this during testing, we set
   * the buffer size on the socket to 4 bytes.  This will ensure that
   * each read and write operation works on at most 4 bytes--a good
   * stress test. */
#if SMALL_LIMITS
#if defined (SO_RCVBUF) && defined (SO_SNDBUF)
  /* Make sure this really is a stream socket (like TCP).  Code using
   * datagram sockets will simply fail miserably if it can never
   * transmit a packet larger than 4 bytes. */
  {
    int sn = sizeof (n);
    if (getsockopt (s, SOL_SOCKET, SO_TYPE, (char *)&n, &sn) < 0
        || n != SOCK_STREAM)
      return;
  }

  n = 4;
  if (setsockopt (s, SOL_SOCKET, SO_RCVBUF, (void *)&n, sizeof (n)) < 0)
    return;
  if (setsockopt (s, SOL_SOCKET, SO_SNDBUF, (void *)&n, sizeof (n)) < 0)
    fatal ("SO_SNDBUF: %s\n", strerror (errno));
#else /* !SO_RCVBUF || !SO_SNDBUF */
#error "Need SO_RCVBUF/SO_SNDBUF for SMALL_LIMITS"
#endif /* SO_RCVBUF && SO_SNDBUF */
#endif /* SMALL_LIMITS */

  /* Enable keepalives to make sockets time out if servers go away. */
  n = 1;
  if (setsockopt (s, SOL_SOCKET, SO_KEEPALIVE, (void *) &n, sizeof (n)) < 0)
    fatal ("SO_KEEPALIVE: %s\n", strerror (errno));
}


int tsocket(int domain, int type, int protocol)

{
	int  s;

	s = socket (domain, type, protocol);

	if (s >= 0) {
		make_async(s);
		fd_open(s);
	}

	return (s);
}

int tconnect(int s, struct sockaddr *name, int namelen)

{
	int ret;
	int ret2;


	ret = connect(s, name, namelen);

	if ((ret != -1) || (errno != EINPROGRESS)) {
		return (ret);
	}

	if ((ret = fd_sleep(s, 1)) < 0)
		return (ret);
	
	{
		int ret_size = sizeof(ret);
		ret2 = getsockopt(s, SOL_SOCKET, SO_ERROR, (void *) &ret, &ret_size);

		if (ret2 < 0) {
			prerror("getsockopt");
			exit (-1);
		}
	}
	      
	if (ret != 0) {
		errno = ret;
		return (-1);
	} 

	return (0);
}


int tbind(int s, const struct sockaddr *name, int namelen)

{
	return (bind(s, name, namelen));
}

int tlisten(int s, int conn)

{
	return (listen(s, conn));
}

int taccept(int s, struct sockaddr *addr, int *addrlen)

{
	int   news;

	while (1) {
		if ((news = fd_sleep(s, 0)) < 0)
			break;

		if ((news = accept(s, addr, addrlen)) >= 0)
			break;

		if (errno != EWOULDBLOCK && errno != EAGAIN)
			break;
	}

	if (news >= 0) {
		make_async(news);
		fd_open(news);
	}

	return (news);
}
  
int tread(int s, void *buf, size_t nbytes)

{
	int ret;

	/* How to deal with zero byte writes? */

	while ((ret = read(s, buf, nbytes)) < 0) {
		if (ret < 0 && errno != EAGAIN)
			break;

		if ((ret = fd_sleep(s, 0)) < 0)
			break;
	}
	
	return (ret);
}

int twrite(int s, void *buf, size_t nbytes)

{
	int ret;
	int bytes_left = nbytes;
	char *bufptr = buf;

	/* How to deal with zero byte writes? */

	while ((ret = write(s, bufptr, bytes_left)) != bytes_left) {
		if (ret < 0 && errno != EAGAIN)
			break;

		if (ret > 0) {
			bufptr += ret;
			bytes_left -= ret;
			if (bytes_left == 0) {
				break;
			}				
		}

		if ((ret = fd_sleep(s, 1)) < 0)
			break;
	}

	if (ret != -1)
		return (nbytes);

	return (ret);
}

int tshutdown (int s, int b)

{
	fd_shutdown(s, b);
	return (shutdown(s, b));
}

int tclose(int s)

{
	fd_shutdown(s, 2);
	return (close (s));
}
